# For digest operations.
import hashlib

# For saving the images from Pillow.
from io import BytesIO

# Pillow for Image processing.
from PIL import Image, ImageDraw

# For decoding hex values (works both for Python 2.7.x and Python 3.x).
import binascii


class Generator(object):
    """
    Factory class that can be used for generating the identicons
    deterministically based on hash of the passed data.

    Resulting identicons are images of requested size with optional padding. The
    identicon (without padding) consists out of M x N blocks, laid out in a
    rectangle, where M is the number of blocks in each column, while N is number
    of blocks in each row.

    Each block is a smallself rectangle on its own, filled using the foreground or
    background colour.

    The foreground is picked randomly, based on the passed data, from the list
    of foreground colours set during initialisation of the generator.

    The blocks are always laid-out in such a way that the identicon will be
    symterical by the Y axis. The center of symetry will be the central column
    of blocks.

    Simply put, the generated identicons are small symmetric mosaics with
    optional padding.
    """

    def __init__(self, rows, columns, digest=hashlib.md5, foreground=["#000000"], background="#ffffff"):
        """
        Initialises an instance of identicon generator. The instance can be used
        for creating identicons with differing image formats, sizes, and with
        different padding.

        Arguments:

          rows - Number of block rows in an identicon.

          columns - Number of block columns in an identicon.

          digest - Digest class that should be used for the user's data. The
          class should support accepting a single constructor argument for
          passing the data on which the digest will be run. Instances of the
          class should also support a single hexdigest() method that should
          return a digest of passed data as a hex string. Default is
          hashlib.md5. Selection of the digest will limit the maximum values
          that can be set for rows and columns. Digest needs to be able to
          generate (columns / 2 + columns % 2) * rows + 8 bits of entropy.

          foreground - List of colours which should be used for drawing the
          identicon. Each element should be a string of format supported by the
          PIL.ImageColor module. Default is ["#000000"] (only black).

          background - Colour (single) which should be used for background and
          padding, represented as a string of format supported by the
          PIL.ImageColor module. Default is "#ffffff" (white).
        """

        # Check if the digest produces sufficient entropy for identicon
        # generation.
        entropy_provided = len(digest(b"test").hexdigest()) // 2 * 8
        entropy_required = (columns // 2 + columns % 2) * rows + 8

        if entropy_provided < entropy_required:
            raise ValueError("Passed digest '%s' is not capable of providing %d bits of entropy" % (str(digest), entropy_required))

        # Set the expected digest size. This is used later on to detect if
        # passed data is a digest already or not.
        self.digest_entropy = entropy_provided

        self.rows = rows
        self.columns = columns

        self.foreground = foreground
        self.background = background

        self.digest = digest

    def _get_bit(self, n, hash_bytes):
        """
        Determines if the n-th bit of passed bytes is 1 or 0.

        Arguments:

          hash_bytes - List of hash byte values for which the n-th bit value
          should be checked. Each element of the list should be an integer from
          0 to 255.

        Returns:

          True if the bit is 1. False if the bit is 0.
        """

        if hash_bytes[n // 8] >> int(8 - ((n % 8) + 1)) & 1 == 1:
            return True

        return False

    def _generate_matrix(self, hash_bytes):
        """
        Generates matrix that describes which blocks should be coloured.

        Arguments:
          hash_bytes - List of hash byte values for which the identicon is being
          generated. Each element of the list should be an integer from 0 to
          255.

        Returns:
          List of rows, where each element in a row is boolean. True means the
          foreground colour should be used, False means a background colour
          should be used.
        """

        # Since the identicon needs to be symmetric, we'll need to work on half
        # the columns (rounded-up), and reflect where necessary.
        half_columns = self.columns // 2 + self.columns % 2
        cells = self.rows * half_columns

        # Initialise the matrix (list of rows) that will be returned.
        matrix = [[False] * self.columns for _ in range(self.rows)]

        # Process the cells one by one.
        for cell in range(cells):

            # If the bit from hash correpsonding to this cell is 1, mark the
            # cell as foreground one. Do not use first byte (since that one is
            # used for determining the foreground colour.
            if self._get_bit(cell, hash_bytes[1:]):

                # Determine the cell coordinates in matrix.
                column = cell // self.columns
                row = cell % self.rows

                # Mark the cell and its reflection. Central column may get
                # marked twice, but we don't care.
                matrix[row][column] = True
                matrix[row][self.columns - column - 1] = True

        return matrix

    def _data_to_digest_byte_list(self, data):
        """
        Creates digest of data, returning it as a list where every element is a
        single byte of digest (an integer between 0 and 255).

        No digest will be calculated on the data if the passed data is already a
        valid hex string representation of digest, and the passed value will be
        used as digest in hex string format instead.

        Arguments:

          data - Raw data or hex string representation of existing digest for
          which a list of one-byte digest values should be returned.

        Returns:

          List of integers where each element is between 0 and 255, and
          repesents a single byte of a data digest.
        """

        # If data seems to provide identical amount of entropy as digest, it
        # could be a hex digest already.
        if len(data) // 2 == self.digest_entropy // 8:
            try:
                binascii.unhexlify(data.encode('utf-8'))
                digest = data.encode('utf-8')
            # Handle Python 2.x exception.
            except (TypeError):
                digest = self.digest(data.encode('utf-8')).hexdigest()
            # Handle Python 3.x exception.
            except (binascii.Error):
                digest = self.digest(data.encode('utf-8')).hexdigest()
        else:
            digest = self.digest(data.encode('utf-8')).hexdigest()

        return [int(digest[i * 2:i * 2 + 2], 16) for i in range(16)]

    def _generate_image(self, matrix, width, height, padding, foreground, background, image_format):
        """
        Generates an identicon image in requested image format out of the passed
        block matrix, with the requested width, height, padding, foreground
        colour, background colour, and image format.

        Arguments:

          matrix - Matrix describing which blocks in the identicon should be
          painted with foreground (background if inverted) colour.

          width - Width of resulting identicon image in pixels.

          height - Height of resulting identicon image in pixels.

          padding - Tuple describing padding around the generated identicon. The
          tuple should consist out of four values, where each value is the
          number of pixels to use for padding. The order in tuple is: top,
          bottom, left, right.

          foreground - Colour which should be used for foreground (filled
          blocks), represented as a string of format supported by the
          PIL.ImageColor module.

          background - Colour which should be used for background and padding,
          represented as a string of format supported by the PIL.ImageColor
          module.

          image_format - Format to use for the image. Format needs to be
          supported by the Pillow library.

        Returns:

          Identicon image in requested format, returned as a byte list.
        """

        # Set-up a new image object, setting the background to provided value.
        image = Image.new("RGBA", (width + padding[2] + padding[3], height + padding[0] + padding[1]), background)

        # Set-up a draw image (for drawing the blocks).
        draw = ImageDraw.Draw(image)

        # Calculate the block widht and height.
        block_width = width // self.columns
        block_height = height // self.rows

        # Go through all the elements of a matrix, and draw the rectangles.
        for row, row_columns in enumerate(matrix):
            for column, cell in enumerate(row_columns):
                if cell:
                    # Set-up the coordinates for a block.
                    x1 = padding[2] + column * block_width
                    y1 = padding[0] + row * block_height
                    x2 = padding[2] + (column + 1) * block_width - 1
                    y2 = padding[0] + (row + 1) * block_height - 1

                    # Draw the rectangle.
                    draw.rectangle((x1, y1, x2, y2), fill=foreground)

        # Set-up a stream where image will be saved.
        stream = BytesIO()

        # Save the image to stream.
        try:
            image.save(stream, format=image_format, optimize=True)
        except KeyError:
            raise ValueError("Pillow does not support requested image format: %s" % image_format)
        image_raw = stream.getvalue()
        stream.close()

        # Return the resulting image.
        return image_raw

    def _generate_ascii(self, matrix, foreground, background):
        """
        Generates an identicon "image" in the ASCII format. The image will just
        output the matrix used to generate the identicon.

        Arguments:

          matrix - Matrix describing which blocks in the identicon should be
          painted with foreground (background if inverted) colour.

          foreground - Character which should be used for representing
          foreground.

          background - Character which should be used for representing
          background.

        Returns:

          ASCII representation of an identicon image, where one block is one
          character.
        """

        return "\n".join(["".join([foreground if cell else background for cell in row]) for row in matrix])

    def generate(self, data, width, height, padding=(0, 0, 0, 0), output_format="png", inverted=False):
        """
        Generates an identicon image with requested width, height, padding, and
        output format, optionally inverting the colours in the indeticon
        (swapping background and foreground colours) if requested.

        Arguments:

          data - Hashed or raw data that will be used for generating the
          identicon.

          width - Width of resulting identicon image in pixels.

          height - Height of resulting identicon image in pixels.

          padding - Tuple describing padding around the generated identicon. The
          tuple should consist out of four values, where each value is the
          number of pixels to use for padding. The order in tuple is: top,
          bottom, left, right.

          output_format - Output format of resulting identicon image. Supported
          formats are anything that is supported by Pillow, plus a special
          "ascii" mode.

          inverted - Specifies whether the block colours should be inverted or
          not. Default is False.

        Returns:

          Byte representation of an identicon image.
        """

        # Calculate the digest, and get byte list.
        digest_byte_list = self._data_to_digest_byte_list(data)

        # Create the matrix describing which block should be filled-in.
        matrix = self._generate_matrix(digest_byte_list)

        # Determine the background and foreground colours.
        if output_format == "ascii":
            foreground = "+"
            background = "-"
        else:
            background = self.background
            foreground = self.foreground[digest_byte_list[0] % len(self.foreground)]

        # Swtich the colours if inverted image was requested.
        if inverted:
            foreground, background = background, foreground

        # Generate the identicon in requested format.
        if output_format == "ascii":
            return self._generate_ascii(matrix, foreground, background)
        else:
            return self._generate_image(matrix, width, height, padding, foreground, background, output_format)



# Standard library imports.
import hashlib
import unittest
from io import BytesIO

# Third-party Python library imports.
import mock
import PIL
import PIL.ImageChops

# Library imports.
from pydenticon import Generator


class GeneratorTest(unittest.TestCase):
    """
    Implements tests for pydenticon.Generator class.
    """

    def test_init_entropy(self):
        """
        Tests if the constructor properly checks for entropy provided by a
        digest algorithm.
        """

        # Set-up the mock instance.
        hexdigest_method = mock.MagicMock(return_value="aabb")
        digest_instance = mock.MagicMock()
        digest_instance.hexdigest = hexdigest_method

        # Set-up digest function that will always return the same digest
        # instance.
        digest_method = mock.MagicMock(return_value=digest_instance)

        # This should require 23 bits of entropy, while the digest we defined
        # provided 2*8 bits of entropy (2 bytes).
        self.assertRaises(ValueError, Generator, 5, 5, digest=digest_method)

    def test_init_parameters(self):
        """
        Verifies that the constructor sets-up the instance properties correctly.
        """

        generator = Generator(5, 5, digest=hashlib.sha1, foreground=["#111111", "#222222"], background="#aabbcc")

        # sha1 provides 160 bits of entropy - 20 bytes.
        self.assertEqual(generator.digest_entropy, 20 * 8)
        self.assertEqual(generator.digest, hashlib.sha1)
        self.assertEqual(generator.rows, 5)
        self.assertEqual(generator.columns, 5)
        self.assertEqual(generator.foreground, ["#111111", "#222222"])
        self.assertEqual(generator.background, "#aabbcc")

    def test_get_bit(self):
        """
        Tests if the check whether bit is 1 or 0 is performed correctly.
        """

        generator = Generator(5, 5)
        hash_bytes = [0b10010001, 0b10001000, 0b00111001]

        # Check a couple of bits from the above hash bytes.
        self.assertEqual(True, generator._get_bit(0, hash_bytes))
        self.assertEqual(True, generator._get_bit(7, hash_bytes))
        self.assertEqual(False, generator._get_bit(22, hash_bytes))
        self.assertEqual(True, generator._get_bit(23, hash_bytes))

    def test_generate_matrix(self):
        """
        Verifies that the matrix is generated correctly based on passed hashed
        bytes.
        """

        # The resulting half-matrix should be as follows (first byte is for
        # ignored in matrix generation):
        #
        # 100
        # 011
        # 100
        # 001
        # 110
        hash_bytes = [0b11111111, 0b10101010, 0b01010101]

        expected_matrix = [
            [True, False, False, False, True],
            [False, True, True, True, False],
            [True, False, False, False, True],
            [False, False, True, False, False],
            [True, True, False, True, True],
            ]

        generator = Generator(5, 5)

        matrix = generator._generate_matrix(hash_bytes)

        self.assertEqual(matrix, expected_matrix)

    def test_data_to_digest_byte_list_raw(self):
        """
        Test if correct digest byte list is returned for raw (non-hex-digest)
        data passed to the method.
        """

        # Set-up some raw data, and set-up the expected result.
        data = "this is a test\n"
        expected_digest_byte_list = [225, 156, 18, 131, 201, 37, 179, 32, 102, 133, 255, 82, 42, 207, 227, 230]

        # Instantiate a generator.
        generator = Generator(5, 5, digest=hashlib.md5)

        # Call the method and get the results.
        digest_byte_list = generator._data_to_digest_byte_list(data)

        # Verify the expected and actual result are identical.
        self.assertEqual(expected_digest_byte_list, digest_byte_list)

    def test_data_to_digest_byte_list_hex(self):
        """
        Test if correct digest byte list is returned for passed hex digest
        string.
        """

        # Set-up some test hex digest (md5), and expected result.
        hex_digest = "e19c1283c925b3206685ff522acfe3e6"
        expected_digest_byte_list = [225, 156, 18, 131, 201, 37, 179, 32, 102, 133, 255, 82, 42, 207, 227, 230]

        # Instantiate a generator.
        generator = Generator(5, 5, digest=hashlib.md5)

        # Call the method and get the results.
        digest_byte_list = generator._data_to_digest_byte_list(hex_digest)

        # Verify the expected and actual result are identical.
        self.assertEqual(expected_digest_byte_list, digest_byte_list)

    def test_data_to_digest_byte_list_hex_lookalike(self):
        """
        Test if correct digest byte list is returned for passed raw data that
        has same length as hex digest string.
        """

        # Set-up some test hex digest (md5), and expected result.
        data = "qqwweerrttyyuuiiooppaassddffgghh"
        expected_digest_byte_list = [25, 182, 52, 218, 118, 220, 26, 145, 164, 222, 33, 221, 183, 140, 98, 246]

        # Instantiate a generator.
        generator = Generator(5, 5, digest=hashlib.md5)

        # Call the method and get the results.
        digest_byte_list = generator._data_to_digest_byte_list(data)

        # Verify the expected and actual result are identical.
        self.assertEqual(expected_digest_byte_list, digest_byte_list)

    def test_generate_image_basics(self):
        """
        Tests some basics about generated PNG identicon image. This includes:

        - Dimensions of generated image.
        - Format of generated image.
        - Mode of generated image.
        """

        # Set-up parameters that will be used for generating the image.
        width = 200
        height = 200
        padding = [20, 20, 20, 20]
        foreground = "#ffffff"
        background = "#000000"
        matrix = [
            [0, 0, 1, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 1, 1, 1, 0],
            [0, 1, 1, 1, 0],
            ]

        # Set-up a generator.
        generator = Generator(5, 5)

        # Generate the raw image.
        raw_image = generator._generate_image(matrix, width, height, padding, foreground, background, "png")

        # Try to load the raw image.
        image_stream = BytesIO(raw_image)
        image = PIL.Image.open(image_stream)

        # Verify image size, format, and mode.
        self.assertEqual(image.size[0], 240)
        self.assertEqual(image.size[1], 240)
        self.assertEqual(image.format, "PNG")
        self.assertEqual(image.mode, "RGBA")

    def test_generate_ascii(self):
        """
        Tests the generated identicon in ASCII format.
        """

        # Set-up parameters that will be used for generating the image.
        foreground = "1"
        background = "0"
        matrix = [
            [0, 0, 1, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 1, 1, 1, 0],
            [0, 1, 1, 1, 0],
            ]

        # Set-up a generator.
        generator = Generator(5, 5)

        # Generate the ASCII image.
        ascii_image = generator._generate_ascii(matrix, foreground, background)

        # Verify that the result is as expected.
        expected_result = """00100
00100
00100
01110
01110"""
        self.assertEqual(ascii_image, expected_result)

    def test_generate_format(self):
        """
        Tests if identicons are generated in requested format.
        """

        # Set-up a generator.
        generator = Generator(5, 5)

        # Set-up some test data.
        data = "some test data"

        # Verify that PNG image is returned when requested.
        raw_image = generator.generate(data, 200, 200, output_format="png")
        image_stream = BytesIO(raw_image)
        image = PIL.Image.open(image_stream)
        self.assertEqual(image.format, "PNG")

        # Verify that JPEG image is returned when requested.
        raw_image = generator.generate(data, 200, 200, output_format="jpeg")
        image_stream = BytesIO(raw_image)
        image = PIL.Image.open(image_stream)
        self.assertEqual(image.format, "JPEG")

        # Verify that GIF image is returned when requested.
        raw_image = generator.generate(data, 200, 200, output_format="gif")
        image_stream = BytesIO(raw_image)
        image = PIL.Image.open(image_stream)
        self.assertEqual(image.format, "GIF")

        # Verify that ASCII "image" is returned when requested.
        raw_image = generator.generate(data, 200, 200, output_format="ascii")
        self.assertIsInstance(raw_image, str)

    def test_generate_format_invalid(self):
        """
        Tests if an exception is raised in case an unsupported format is
        requested when generating the identicon.
        """

        # Set-up a generator.
        generator = Generator(5, 5)

        # Set-up some test data.
        data = "some test data"

        # Verify that an exception is raised in case of unsupported format.
        self.assertRaises(ValueError, generator.generate, data, 200, 200, output_format="invalid")

    @mock.patch.object(Generator, '_generate_image')
    def test_generate_inverted_png(self, generate_image_mock):
        """
        Tests if the foreground and background are properly inverted when
        generating PNG images.
        """

        # Set-up some test data.
        data = "Some test data"

        # Set-up one foreground and background colour.
        foreground = "#ffffff"
        background = "#000000"

        # Set-up the generator.
        generator = Generator(5, 5, foreground=[foreground], background=background)

        # Verify that colours are picked correctly when no inverstion is requsted.
        generator.generate(data, 200, 200, inverted=False, output_format="png")
        generate_image_mock.assert_called_with(mock.ANY, mock.ANY, mock.ANY, mock.ANY, foreground, background, "png")

        # Verify that colours are picked correctly when inversion is requsted.
        generator.generate(data, 200, 200, inverted=True, output_format="png")
        generate_image_mock.assert_called_with(mock.ANY, mock.ANY, mock.ANY, mock.ANY, background, foreground, "png")

    @mock.patch.object(Generator, '_generate_ascii')
    def test_generate_inverted_ascii(self, generate_ascii_mock):
        """
        Tests if the foreground and background are properly inverted when
        generating ASCII "images".
        """

        # Set-up some test data.
        data = "Some test data"

        # Set-up one foreground and background colour. These are not used for
        # ASCII itself (instead a plus/minus sign is used).
        foreground = "#ffffff"
        background = "#000000"

        # Set-up the generator.
        generator = Generator(5, 5, foreground=[foreground], background=background)

        # Verify that foreground/background is picked correctly when no
        # inverstion is requsted.
        generator.generate(data, 200, 200, inverted=False, output_format="ascii")
        generate_ascii_mock.assert_called_with(mock.ANY, "+", "-")

        # Verify that foreground/background is picked correctly when inversion
        # is requsted.
        generator.generate(data, 200, 200, inverted=True, output_format="ascii")
        generate_ascii_mock.assert_called_with(mock.ANY, "-", "+")

    @mock.patch.object(Generator, '_generate_image')
    def test_generate_foreground(self, generate_image_mock):
        """
        Tests if the foreground colour is picked correctly.
        """

        # Set-up some foreground colours and a single background colour.
        foreground = ["#000000", "#111111", "#222222", "#333333", "#444444", "#555555"]
        background = "#ffffff"

        # Set-up the generator.
        generator = Generator(5, 5, foreground=foreground, background=background)

        # The first byte of hex digest should be 121 for this data, which should
        # result in foreground colour of index '1'.
        data = "some test data"
        generator.generate(data, 200, 200)
        generate_image_mock.assert_called_with(mock.ANY, mock.ANY, mock.ANY, mock.ANY, foreground[1], background, "png")

        # The first byte of hex digest should be 149 for this data, which should
        # result in foreground colour of index '5'.
        data = "some other test data"
        generator.generate(data, 200, 200)
        generate_image_mock.assert_called_with(mock.ANY, mock.ANY, mock.ANY, mock.ANY, foreground[5], background, "png")

    def test_generate_image_compare(self):
        """
        Tests generated PNG identicon against a set of pre-generated samples.
        """

        # Set-up a list of foreground colours (taken from Sigil). Same as used
        # for reference images.
        foreground = ["rgb(45,79,255)",
                      "rgb(254,180,44)",
                      "rgb(226,121,234)",
                      "rgb(30,179,253)",
                      "rgb(232,77,65)",
                      "rgb(49,203,115)",
                      "rgb(141,69,170)"]

        # Set-up a background colour (taken from Sigil). Same as used for
        # reference images.
        background = "rgb(224,224,224)"

        # Set-up parameters equivalent as used for samples.
        width = 200
        height = 200
        padding = (20, 20, 20, 20)

        # Load the reference images, making sure they're in RGBA mode.
        test1_ref = PIL.Image.open("tests/samples/test1.png").convert(mode="RGBA")
        test2_ref = PIL.Image.open("tests/samples/test2.png").convert(mode="RGBA")
        test3_ref = PIL.Image.open("tests/samples/test3.png").convert(mode="RGBA")

        # Set-up the Generator.
        generator = Generator(5, 5, foreground=foreground, background=background)

        # Generate first test identicon.
        raw_image = generator.generate("test1", width, height, padding=padding)
        image_stream = BytesIO(raw_image)
        test1 = PIL.Image.open(image_stream)

        # Generate second test identicon.
        raw_image = generator.generate("test2", width, height, padding=padding)
        image_stream = BytesIO(raw_image)
        test2 = PIL.Image.open(image_stream)

        # Generate third test identicon.
        raw_image = generator.generate("test3", width, height, padding=padding)
        image_stream = BytesIO(raw_image)
        test3 = PIL.Image.open(image_stream)

        # Calculate differences between generated identicons and references.
        diff1 = PIL.ImageChops.difference(test1, test1_ref)
        diff2 = PIL.ImageChops.difference(test2, test2_ref)
        diff3 = PIL.ImageChops.difference(test3, test3_ref)

        # Verify that all the diffs are essentially black (i.e. no differences
        # between generated identicons and reference samples).
        expected_extrema = ((0, 0), (0, 0), (0, 0), (0, 0))

        self.assertEqual(diff1.getextrema(), expected_extrema)
        self.assertEqual(diff2.getextrema(), expected_extrema)
        self.assertEqual(diff3.getextrema(), expected_extrema)

if __name__ == '__main__':
    unittest.main()



import os
from setuptools import setup

README = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()
INSTALL_REQUIREMENTS = ["Pillow"]
TEST_REQUIREMENTS = ["mock"]

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='pydenticon',
    version='0.3',
    packages=['pydenticon'],
    include_package_data=True,
    license='BSD',  # example license
    description='Library for generating identicons. Port of Sigil (https://github.com/cupcake/sigil) with enhancements.',
    long_description=README,
    url='https://github.com/azaghal/pydenticon',
    author='Branko Majic',
    author_email='branko@majic.rs',
    install_requires=INSTALL_REQUIREMENTS,
    tests_require=TEST_REQUIREMENTS,
    test_suite="tests",
    classifiers=[
        'Environment :: Other Environment',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Software Development :: Libraries',
    ],
)
